#!/usr/bin/perl
###############################################################################
# Description:
#       General purpose utility function for providing JSON or XML
#       interface to server-side database tables.  Should be compatible
#       with ExtJS (tested with JSON) and Flex (tested with XML).
#
#       Server-side configuration defines the SQL.  Sessions are managed
#       with CGI::Session.  A pluggable authentication approach allows for
#       different login rules.
#
#       This contains the top-level "Main" body of the Jarvis processing,
#       call it either from Agent.pm (for mod_perl) or agent.pl (if not
#       using mod_perl).
#
# Licence:
#       This file is part of the Jarvis WebApp/Database gateway utility.
#
#       Jarvis is free software: you can redistribute it and/or modify
#       it under the terms of the GNU General Public License as published by
#       the Free Software Foundation, either version 3 of the License, or
#       (at your option) any later version.
#
#       Jarvis is distributed in the hope that it will be useful,
#       but WITHOUT ANY WARRANTY; without even the implied warranty of
#       MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#       GNU General Public License for more details.
#
#       You should have received a copy of the GNU General Public License
#       along with Jarvis.  If not, see <http://www.gnu.org/licenses/>.
#
#       This software is Copyright 2008 by Jonathan Couper-Smartt.
###############################################################################
use strict;
use warnings;

package Jarvis::Main;

use CGI;
use Data::Dumper;
use Encode 'decode_utf8';
use File::Basename;

use Jarvis::Error;
use Jarvis::Config;
use Jarvis::Login;
use Jarvis::Dataset;
use Jarvis::Status;
use Jarvis::Habitat;
use Jarvis::Exec;
use Jarvis::Plugin;
use Jarvis::Hook;
use Jarvis::DB;
use Jarvis::Route;

###############################################################################
# Global variables.
###############################################################################
#
# Note that global variables under mod_perl require careful consideration!
#
# Specifically, you must ensure that all variables which require
# re-initialisation for each invocation will receive it.
#

# This is our CGI object.
# We pass it into our Jasper::Config, and also use it in our "die" error handler.
#
# It is safe because it is re-initialised in Main::do.
#
my $cgi = undef;

# This is our Jasper Config object which is passed around everywhere, and is
# used in our "die" error handler so it needs to be module wide.
#
# It is safe because it is re-initialised in Main::do.
#
my $jconfig = undef;

# This is where we might look for etc directories.
#
# It is safe because it is never modified so does not need re-initialisation.
#
my @etc = ('/etc/jarvis', '/opt/jarvis/etc');

# Version 6.0.0.
$Jarvis::Main::JARVIS_VERSION = 700;

###############################################################################
# Generate a random UUID. Avoid any reliance on 3rd party UUID generator
# perl modules due to the difficulty in yum/apt installation procedures on
# most target systems.
#
# This approach here is a UUID generated based on random numbers - and is based
# off of Perl's UUID::Tiny module.
###############################################################################
#
sub generate_uuid {
    my $uuid = '';
    for (1 .. 4) {
        my $top = int(rand(65536)) % 65536;
        my $bottom = int(rand(65536)) % 65536;
        $uuid .= pack 'I', (($top << 16) | $bottom);
    }
    substr $uuid, 6, 1, chr(ord(substr($uuid, 6, 1)) & 0x0f);
    substr $uuid, 8, 1, chr(ord(substr $uuid, 8, 1) & 0x3f | 0x80);
    return join '-', map { unpack 'H*', $_ } map { substr $uuid, 0, $_, '' } ( 4, 2, 2, 2, 6 );
}

###############################################################################
# Setup error handler.
###############################################################################
#
sub error_handler {
    my ($msg) = @_;

    # Are we parsing ($^S undef) or executing ($^S == 1) an eval?
    #
    # If this is an "eval" within a Jarvis plugin, then we don't want to fire
    # this "die" handler and shut down the whole process, that's a massive
    # over-reaction.
    #
    # The problem arises when running under mod_perl.  The mod_perl wrapper
    # runs everything inside its own "eval", which we detect, and confuse
    # with a Jarvis plugin "eval".
    #
    if ((! defined $^S) or ($^S == 1)) {

    	# Go up the caller stack and see if we find an "(eval)" before we hit "Jarvis::Main::do".
     	my $frames = 0;
    	while (1) {
    	    my @frame = caller ($frames);
    	    if (! scalar (@frame)) {
                die "Cannot find '(eval)' frame in call stack.";
    	    }
    	    my $subroutine = $frame[3];

    	    # If we hit "(eval)" before Jarvis::Main::do, then that means this is a user-eval.
    	    # We don't want to die at all in this case!  Just return and let the user's "eval"
    	    # post-processing run and deal with things.
    	    if ($subroutine eq '(eval)') {
                return;

    	    # Otherwise if we got to "Jarvis::Main::do" without hitting a user "(eval)" then
    	    # this means that something really did call a "die", and we really do want to
    	    # invoke this handler and return a message back to the client.
    	    } elsif ($subroutine eq 'Jarvis::Main::do') {
                last;
    	    }
    	    $frames++;
    	}
    }

    # Truncate any thing after a null-TERM.  This is because LDAP error
    # messages sometimes put some junk in, which means that the browser
    # thinks the error is binary, and doesn't display it.
    $msg =~ s/\x00.*$//;
    $msg =~ s/\s*$/\n/;

    # Return error to client.  Note that we do not print stack trace to user,
    # since that is a potential security weakness.
    $jconfig->{status} = $jconfig->{status} || "500 Internal Server Error";
    my $status = $jconfig->{status};
    &Jarvis::Config::add_http_headers ($jconfig,{ -status => $status, -type => "text/plain; charset=UTF-8", 'Content-Disposition' => "inline; filename=error.txt" });
    &Jarvis::Error::print_utf8 (\$cgi->header ($jconfig->{http_headers}));
    my $output = $msg;
    if ($status =~ /^500/) {
        $output = &Jarvis::Error::format_message ($jconfig, $jconfig->{error_response_format} || "[%T][%R] %M", 'fatal', $msg);
    }
    &Jarvis::Error::print_utf8 (\$output);

    # Print to error log.  Include stack trace if debug is enabled.
    if ($status =~ /^500/) {
        my $long_msg = &Jarvis::Error::format_message ($jconfig, $jconfig->{log_format}, 'fatal', $msg);

        # Include the URI.
        if (! $jconfig->{debug}) {
            $long_msg = $long_msg . "        URI = $ENV{REQUEST_URI}";
        }

        # The message may contain UTF-8 content, not sure.  Better to be safe.
        $long_msg = $jconfig->{debug} ? Carp::longmess $long_msg : Carp::shortmess $long_msg;
        &Jarvis::Error::print_utf8 (\*STDERR, \$long_msg);
    }

    # We MUST ensure that ALL the cached database handles are removed.
    # Otherwise, under mod_perl, the next application would get OUR database handles!
    &Jarvis::DB::disconnect ($jconfig, undef, undef, 1);

    # Under mod_perl this will be ModPerl::Util::exit (), which won't really end the process.
    # Under non-mod_perl, this will really exit the process.
    exit ();
}

###############################################################################
# Get $jconfig singleton.  Not designed for internal use.
###############################################################################
#
sub jconfig {
    return $jconfig;
}

###############################################################################
# Perform CSRF protection checks for any non global request.
###############################################################################
#
sub check_csrf_protection {

    my ($jconfig, $allowed_groups) = @_;

    # If we have CSRF Protection enabled validate requests that do no have global '**' access.
    if ($jconfig->{csrf_protection} && $allowed_groups ne '**') {
        # Load the CSRF header from the request.
        my $token  = $cgi->http($jconfig->{csrf_header});

        # Load the stored session CSRF token.
        my $session_token = $jconfig->{session}->param ('csrf_token');

        # If the stored session token and the provided token do not match then do not continue the request.
        if (!defined $token || !defined $session_token || $token ne $session_token) {
            $jconfig->{status} = "401 Unauthorized";
            die "Unauthorized Request.\n";
        }
    }
}


###############################################################################
# Main "do" method.
#
# This method is called by either:
#	Main::Agent (mod_perl case) or
#	agent.pl (non-mod_perl case)
###############################################################################
#
sub do {
    my $options = shift;

    $SIG{__WARN__} = sub { die shift };
    $SIG{__DIE__} = \&Jarvis::Main::error_handler;

    # Optional mod-perl stream output variable
    my $mod_perl_io = $options && $options->{mod_perl_io};

    # If we have a CGI parameters configuration file require it now to
    # set any defined CGI parameters.
    foreach my $etc (@etc) {
        if (-f "$etc/cgi_params.pm") {
            require "$etc/cgi_params.pm";
        }
    }

    # CGI object for all sorts of things.
    $cgi = ($options && $options->{cgi}) || new CGI;

    # Check that the CGI does not exceed any globally configured parameters.
    if (defined $cgi->cgi_error()) {
        $jconfig->{status} = $cgi->cgi_error();
        die  "\n";
    }

    # Perl CGI does not explicitly catch when file downloads are disabled. This causes
    # Jarvis to die when the CGI object is not fully initialized. We need to catch this
    # explicit case before we continue with any processing.
    if ($CGI::DISABLE_UPLOADS) {
        if (grep $cgi->param(), 'some_file') {
            $jconfig->{status} = "400 Bad Request";
            die "\n";
        }
    }

    # Environment variables.
    my $jarvis_root = $ENV {JARVIS_ROOT};
    foreach my $inc (@INC) {
        last if $jarvis_root;
        if (-f "$inc/Jarvis/Main.pm") {
            $jarvis_root = dirname ($inc);
        }
    }
    $jarvis_root || die "Cannot determine JARVIS_ROOT.";

    my $jarvis_etc = $ENV {JARVIS_ETC};
    foreach my $etc (@etc) {
        last if $jarvis_etc;
        if (-d $etc) {
            $jarvis_etc = $etc;
        }
    }
    $jarvis_etc || die "Cannot determine JARVIS_ETC.";

    ###############################################################################
    # Check basic HTML parameters.
    ###############################################################################
    #
    my $script_name = $cgi->script_name();
    my $path = $cgi->path_info() ||
        die "Missing path info.  Send $script_name/<app-name>/<arg0>[/<arg1>...] in URI!\n";

    my $method = $cgi->request_method() || 'GET';
    #
    # We don't suppport OPTIONS requests, so return a 501 Not Implemented.
    if ($method eq "OPTIONS") {
        &Jarvis::Config::add_http_headers($jconfig, {-status => '501 Not Implemented', -type => "text/plain", 'Content-Disposition' => "inline; filename=error.txt"});
        print $cgi->header($jconfig->{http_headers});
        return;
    }

    ###############################################################################
    # Some additional parameter parsing code, because of CGI.pm oddness.
    ###############################################################################
    #
    # Explanation: In the heart of CGI.pm, there is special handling for content
    # type "application/xml" which reads the query string args from the URI
    # such as "_method=<transaction-type>" and parses them as CGI parameters.
    #
    # That code is invoked specifically only for POST "application/xml".  However
    # we really want exactly the same done for OTHER application types.  Hence
    # the following.
    #
    my $content_type = $ENV{CONTENT_TYPE} || 'text/plain';

    if (($method eq "POST") && ($content_type ne 'application/xml')) {
        my $query_string = '';
        if (exists $ENV{MOD_PERL}) {
            $query_string = $cgi->r->args;

        } else {
            $query_string = $ENV{QUERY_STRING} if defined $ENV{QUERY_STRING};
            $query_string ||= $ENV{REDIRECT_QUERY_STRING} if defined $ENV{REDIRECT_QUERY_STRING};
        }

        if ($query_string) {
            if ($query_string =~ /[&=;]/) {
                $cgi->parse_params($query_string);
            } else {
                $cgi->add_parameter('keywords');
                $cgi->{keywords} = [$cgi->parse_keywordlist($query_string)];
            }
        }
    }

    ###############################################################################
    # Get our app name and read our $jconfig at last!  Debug can start too.
    ###############################################################################

    # Clean up our path to remove & args, # names.
    $path =~ s|(?<!\\)&.*$||;
    $path =~ s|(?<!\\)#.*$||;

    # Remove leading slash to expose the application name.
    $path =~ s|^/||;

    # Parse our app-name and REST args.  Note that path is no longer
    # URL-encoded by the time it gets to us.  Setting AllowEncodedSlashes doesn't help us
    # get slashes through to this point.  So we do a special case and allow \/ to escape
    # a slash through to our REST args.
    #
    my ($app_name, @path_parts) = split ( m|(?<!\\)/|, $path, -1);
    @path_parts = map { s|\\/|/|g; $_ } @path_parts;

    if (! $app_name) {
        die "Missing app-name.  Send $script_name/<app-name>/<arg0>[/<arg1>...] in URI!\n";
    }
    $app_name =~ m|^[\w\-]+$| || die "Invalid app_name '$app_name'!\n";

    # Create $jconfig object.  This is used everywhere in Jarvis.
    $jconfig = new Jarvis::Config ($app_name, ('etc_dir' => "$jarvis_etc", 'cgi' => $cgi, 'mod_perl_io' => $mod_perl_io ) );

    # Determine client's IP.
    $jconfig->{client_ip} = $ENV{"HTTP_X_FORWARDED_FOR"} || $ENV{"HTTP_CLIENT_IP"} || $ENV{"REMOTE_ADDR"} || '';

    # Determine a unique request ID for this request. Useful for tracing and auditing.
    # This will later get expanded to incorporate the session ID, if we have one.
    $jconfig->{request_id} = generate_uuid();

    # Debug can now occur, since we have called Config!
    &Jarvis::Error::debug ($jconfig, "URI = $ENV{REQUEST_URI}");
    &Jarvis::Error::debug ($jconfig, "Base Path = '$path'.");
    &Jarvis::Error::debug ($jconfig, "App Name = '$app_name'.");
    foreach my $i (0 .. $#path_parts) {
        &Jarvis::Error::debug ($jconfig, "Parsed Path Part $i => '%s'.", $path_parts[$i]);
    }

    ###############################################################################
    # Now we can start with the real action.  Start parsing
    ###############################################################################

    # Now parse the rest of the args and apply our router.  This gives us dataset name too.
    my ($dataset_name, $user_args, $presentation) = &Jarvis::Route::find ($jconfig, \@path_parts);

    # Store this for debugging.
    $jconfig->{dataset_name} = $dataset_name;
    &Jarvis::Error::debug ($jconfig, "Dataset Name = '%s'.", $dataset_name);
    &Jarvis::Error::debug ($jconfig, "Presentation = '%s'.", $presentation);

    # Store the presentation for later encoding.
    $jconfig->{presentation} = $presentation;

    # Show our rest args.
    foreach my $key (sort (keys %$user_args)) {
        &Jarvis::Error::debug ($jconfig, "Rest Arg: '$key' => '%s'.", $$user_args{$key});
    }

    # Merge in the CGI vars.  Do NOT override REST args.
    my $cgi_params = $jconfig->{cgi}->Vars;
    foreach my $name (keys %$cgi_params) {
        next if ($name !~ m/^_?[a-z][a-z0-9_\-]*(\[\])?$/i);
        next if ($name eq 'POSTDATA');
        next if (defined $user_args->{$name});

        my $value = decode_utf8 ($cgi_params->{$name});
        if (length ($value) > 256) {
            &Jarvis::Error::debug ($jconfig, "CGI Param: '$name' => (%d bytes).", length ($value));

        } else {
            &Jarvis::Error::debug ($jconfig, "CGI Param: '$name' => '%s'.", $value);
        }

        if ($name =~ /\[\]$/) {
            my $user_arg  = substr($name, 0, -2);
            $user_args->{$user_arg} = [ split(m{\0}, $value) ];
        } else {
            $user_args->{$name} = $value;
        }
    }

    # Dataset name can't be empty.  Also, it can only be normal characters
    # with "-", and "." for directory separator.
    #
    # Note that we don't check yet for leading and trailing dot and other file
    # security stuff.  We'll do that when we actually go to open the file,
    # because maybe some execs/plugins might allow it, and we don't want
    # to restrict them.
    #
    if ((! defined $dataset_name) || ($dataset_name eq '')) {
        die "All requests require $script_name/$app_name/<dataset-or-special>[/<arg1>...] in URI!\n";
    }
    ($dataset_name =~ m|^[\w\-\.]+$|) || die "Invalid dataset_name '$dataset_name'!\n";

    ###############################################################################
    # Action: "status", "habitat", "logout", "fetch", "update",  or custom
    #           action from Exec or Plugin.
    ###############################################################################
    #
    my $method_param = $jconfig->{method_param};
    if ($method_param) {
        my $new_method = $cgi->param($method_param);
        if ($new_method) {
            &Jarvis::Error::debug ($jconfig, "Using Method '$new_method' instead of '" . $method ."'");
            $method = $new_method;
        }
    }

    my $action = lc ($method) || die "Missing request method!\n";
    ($action =~ m/^\w+$/) || die "Invalid characters in parameter 'action'\n";

    # Now canonicalise our action.
    if ($action eq 'get')    { $action = 'select' };
    if ($action eq 'fetch')  { $action = 'select' };
    if ($action eq 'post')   { $action = 'insert' };
    if ($action eq 'create') { $action = 'insert' };
    if ($action eq 'put')    { $action = 'update' };
    if ($action eq 'patch')  { $action = 'merge'  };

    $jconfig->{action} = $action;

    # Load/Start application-specific start hook(s).
    &Jarvis::Hook::load_global ($jconfig);

    # Login as required.
    &Jarvis::Login::check ($jconfig);

    # Now we have some login data - lets expand our request UUID to be a unique session/request UUID
    &Jarvis::Error::debug ($jconfig, "User Name = '" . $jconfig->{username} . "'");
    &Jarvis::Error::debug ($jconfig, "Group List = '" . $jconfig->{group_list} . "'");
    &Jarvis::Error::debug ($jconfig, "Logged In = " . $jconfig->{logged_in});
    &Jarvis::Error::debug ($jconfig, "Request ID = '" . $jconfig->{request_id} . "'.");
    &Jarvis::Error::debug ($jconfig, "Error String = '" . $jconfig->{error_string} . "'");
    &Jarvis::Error::debug ($jconfig, "Method = '" . $method . "'");
    &Jarvis::Error::debug ($jconfig, "Action = '" . $action . "'");

    # If we have CSRF Protection enabled check that the origin and target locations match for all requests.
    if ($jconfig->{cross_origin_protection}) {
        # If the Origin header is present verify it matches the target origin.
        # Use the user specified session domain when comparing the referer or origin addresses.
        my $host = $jconfig->{scookie_domain};
        if (defined ($ENV{HTTP_ORIGIN})) {
            my $raw_origin = ($ENV{HTTP_ORIGIN} =~ /^(?:.*:\/\/)?(.*?)(?:\/.*)?$/g)[0];
            if ($host ne $raw_origin) {
                &Jarvis::Error::log ($jconfig, "Host HTTP Origin: '$raw_origin' does not match expected host: '$host'");
                $jconfig->{status} = "400 Bad Request";
                die "Target origin does not match window's origin.\n";
            }
        } elsif (defined ($ENV{HTTP_REFERER})) {
            # If the Origin header is not present verify that the HTTP Referrer matches the target origin.
            my $raw_referer = ($ENV{HTTP_REFERER} =~ /^(?:.*:\/\/)?(.*?)(?:\/.*)?$/g)[0];
            if ($host ne $raw_referer) {
                &Jarvis::Error::log ($jconfig, "Host Referer: '$raw_referer' does not match expected host: '$host'");
                $jconfig->{status} = "400 Bad Request";
                die "Target origin does not match window's origin.\n";
            }
        } else {
            # If neither are specified then stop further execution as we cant be sure that we aren't protecting against CSRF.
            &Jarvis::Error::log ($jconfig, "No HTTP_REFERER or HTTP_ORIGIN provided by client.");
            $jconfig->{status} = "400 Bad Request";
            die "Window Origin or Referer not Defined.\n";
        }
    }

    # What kind of dataset?  Used by the tracker only.
    # 's' = sql, 'i' = internal, 'p' = plugin, 'e' = exec, undef for undetermined.
    $jconfig->{dataset_type} = undef;

    # This is the type we use for anything that has structured content JSON or XML.
    my $object_content_type = "text/plain; charset=UTF-8";
    if ($jconfig->{format} =~ /json/i ) {
        if (! $jconfig->{return_json_as_text}) {
            $object_content_type = 'application/json; charset=UTF-8';
        }

    } elsif ($jconfig->{format} =~ /XML/i ) {
        $object_content_type =  'application/xml; charset=UTF-8';
    }

    # All special datasets start with "__".
    #
    # Note that our Plugin and Execs may expect "/appname/<something-else>" so
    # we should be careful not to trample on them.
    #
    # Note that "select" is the only permissible action on special datasets.  We
    # ignore whatever action you supplied.
    #
    if ($dataset_name =~ m/^__/) {
        my $return_text = undef;
        $jconfig->{dataset_type} = 'i';
        $jconfig->{action} = 'select';

        # Status.  I.e. are we logged in?
        if ($dataset_name eq "__status") {
            &Jarvis::Error::debug ($jconfig, "Returning status special dataset.");
            $return_text = &Jarvis::Status::report ($jconfig);

        # Habitat.  Echo the contents of the "<context>...</context>" block in our app-name.xml.
        } elsif ($dataset_name eq "__habitat") {
            &Jarvis::Error::debug ($jconfig, "Returning habitat special dataset.");
            $return_text = &Jarvis::Habitat::print ($jconfig);

        # Logout.  Clear session ID cookie, clean login parameters, then return "logged out" status.
        } elsif ($dataset_name eq "__logout") {
            &Jarvis::Error::debug ($jconfig, "Returning logout special dataset.");
            &Jarvis::Login::logout ($jconfig);
            $return_text = &Jarvis::Status::report ($jconfig);

        # Starts with __ so must be special, but we don't know it.
        } else {
            die "Unknown special dataset '$dataset_name'!\n";
        }

        &Jarvis::Config::add_http_headers ($jconfig, {
            -type => $object_content_type,
            -cookie => $jconfig->{cookie},
            'Cache-Control' => 'no-store, no-cache, must-revalidate'
        });

        # As per the logic for $object_content_type, content type charset is
        # guaranteed to be UTF-8.
        &Jarvis::Error::print_utf8 (\$cgi->header ($jconfig->{http_headers}));
        &Jarvis::Error::print_utf8 (\$return_text);

    # A custom exec for this application?  We hand off entirely for this case,
    # since the MIME type may be special.  Exec::Do will add the cookie in the
    # cases where it is doing the header.  But if the exec script itself is
    # doing all the headers, then there will be no session cookie.
    #
    } elsif (&Jarvis::Exec::do ($jconfig, $dataset_name, $user_args)) {
        # All is well if this returns true.  The action is treated.
        $jconfig->{dataset_type} = 'e';

    # A custom plugin for this application?  This is very similar to an Exec,
    # except that where an exec is a `<command>` system call, a Plugin is a
    # dynamically loaded module method.
    #
    } elsif (&Jarvis::Plugin::do ($jconfig, $dataset_name, $user_args)) {
        # All is well if this returns true.  The action is treated.
        $jconfig->{dataset_type} = 'p';

    # Fetch a regular dataset.
    } elsif ($action eq "select") {
        $jconfig->{dataset_type} = 's';

        my $return_text = &Jarvis::Dataset::fetch ($jconfig, $dataset_name, $user_args);

        #
        # When providing CSV and XLSX output, it is most likely going to be
        # downloaded and stored by users, or downloaded and loaded into a
        # spreadsheet application. So, we suggest it as an attachment, with the
        # filename of the dataset.
        #
        my $is_utf8_content = 0;
        if ($jconfig->{format} eq "csv" || $jconfig->{format} eq "xlsx") {
            my $filename = $jconfig->{return_filename} || $dataset_name . ".$jconfig->{format}";
            $filename =~ s/"/\\"/g;

            if ($jconfig->{format} eq "csv") {
                $is_utf8_content = 1;
                $object_content_type = 'text/csv; charset=UTF-8; name="' . $filename . '"';
            } else {
                $object_content_type = 'application/vnd.ms-excel; name="' . $filename . '"';
            }
            &Jarvis::Config::add_http_headers($jconfig, {
                -type => $object_content_type,
                'Content-Disposition' => 'attachment; filename="' . $filename . '"',
                -cookie => $jconfig->{cookie},
                'Cache-Control' => 'no-store, no-cache, must-revalidate'
            });

        } else {
            &Jarvis::Config::add_http_headers($jconfig,  {
                -type => $object_content_type,
                -cookie => $jconfig->{cookie},
                'Cache-Control' => 'no-store, no-cache, must-revalidate'
            });

            # If we have XSRF Protection enabled then we prefix all JSON responses with ")]}',\n" This prevents executable JSON being
            # being passed to JSON clients.
            if ($jconfig->{xsrf_protection} && $jconfig->{format} =~ /^json/) {
                $return_text = ")]}',\n" . $return_text;
            }

            # As per the logic for $object_content_type, content type charset
            # is guaranteed to be UTF-8.
            $is_utf8_content = 1;
        }

        if ($is_utf8_content) {
            &Jarvis::Error::print_utf8(\$cgi->header($jconfig->{http_headers}));
            &Jarvis::Error::print_utf8(\$return_text);
        } else {
            print $cgi->header($jconfig->{http_headers});
            print $return_text;
        }

    # Modify a regular dataset.
    } elsif (($action eq "insert") || ($action eq "update") || ($action eq "delete") || ($action eq "mixed") || ($action eq "merge")) {
        $jconfig->{dataset_type} = 's';
        my $return_text = &Jarvis::Dataset::store ($jconfig, $dataset_name, $user_args);
        &Jarvis::Config::add_http_headers($jconfig, {-type => "text/plain; charset=UTF-8", -cookie => $jconfig->{cookie}});

        # Content type charset is guaranteed to be UTF-8.
        &Jarvis::Error::print_utf8(\$cgi->header($jconfig->{http_headers}));
        &Jarvis::Error::print_utf8(\$return_text);

    # It's the end of the world as we know it.
    } else {
        &Jarvis::Config::add_http_headers($jconfig, {-status => '501 Not Implemented', -type => "text/plain", 'Content-Disposition' => "inline; filename=error.txt"});
        print $cgi->header($jconfig->{http_headers});
        print "The $method method (action $action) is not recognised by Jarvis.";
    }

    ###############################################################################
    # Cleanup.
    ###############################################################################

    # Unload our global hooks.  This will call ::finish on them.
    &Jarvis::Hook::unload_global ($jconfig);

    # We MUST ensure that ALL the cached database handles are removed.
    # Otherwise, under mod_perl, the next application would get OUR database handles!
    #
    # Note that deep in the internals of mod_perl, the underlying handle may be
    # cached for potential re-use by DBI.  But that will ensure that the password
    # and other information matches before allowing it.
    &Jarvis::DB::disconnect ($jconfig);
}

1;
